package com.pnlyy.midi_support;

import android.media.midi.MidiDevice;
import android.media.midi.MidiDeviceInfo;
import android.media.midi.MidiDeviceStatus;
import android.media.midi.MidiManager;
import android.media.midi.MidiOutputPort;
import android.media.midi.MidiReceiver;
import android.os.Build;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;

import androidx.annotation.NonNull;
import androidx.annotation.RequiresApi;

import com.pnlyy.midi_support.interfaces.OnMidiData;
import com.pnlyy.midi_support.utils.LogTool;

import java.io.IOException;

/**
 * Android M+ midi usb设备支持 by MidiManager
 *
 * @author liben
 * @date 11/30/20
 */
@RequiresApi(api = Build.VERSION_CODES.M)
class MidiSupportAboveM extends MidiManager.DeviceCallback implements MidiSupportImpl<MidiPortWrapper> {

    private final int mType = MidiDeviceInfo.PortInfo.TYPE_OUTPUT;

    protected MidiManager mMidiManager;
    private MidiPortWrapper mCurrentWrapper;

    private MidiOutputPort mOutputPort;
    private MidiDevice mOpenDevice;
    Handler mainHandler = new Handler(Looper.getMainLooper());

    @NonNull
    private final MidiReceiver mMidiReceiver;

    @NonNull
    private final MidiDeviceHelper.OnOTGEvent onOTGEvent;

    /**
     * @param midiManager MidiManager
     * @param receiver    数据接收器
     * @param onOTGEvent  事件监听
     */
    public MidiSupportAboveM(
            @NonNull MidiManager midiManager,
            @NonNull OnMidiData receiver,
            @NonNull MidiDeviceHelper.OnOTGEvent onOTGEvent) {
        mMidiManager = midiManager;
        this.onOTGEvent = onOTGEvent;

        mMidiReceiver = new MidiFramer(receiver);
        mMidiManager.registerDeviceCallback(this,
                new Handler(Looper.getMainLooper()));
    }

    @Override
    public void init() {
        MidiDeviceInfo[] infos = mMidiManager.getDevices();
        for (MidiDeviceInfo info : infos) {
            onDeviceAdded(info);
        }
    }

    @Override
    public void onDestroy() {
        disconnect();
        mMidiManager.unregisterDeviceCallback(this);
    }

    private int getInfoPortCount(final MidiDeviceInfo info) {
        return info.getOutputPortCount();
    }

    @Override
    public void onDeviceAdded(final MidiDeviceInfo info) {
        int portCount = getInfoPortCount(info);
        for (int i = 0; i < portCount; ++i) {
            mCurrentWrapper = new MidiPortWrapper(info, mType, i);
            LogTool.i(mCurrentWrapper + " was added");
            doConnect(mCurrentWrapper);
        }
    }

    @Override
    public void reconnect() {
        MidiPortWrapper wrapper = getCurrentWrapper();
        if (wrapper != null) {
            doConnect(wrapper);
        }
    }

    @Override
    public void onDeviceRemoved(final MidiDeviceInfo info) {
        int portCount = getInfoPortCount(info);
        for (int i = 0; i < portCount; ++i) {
            MidiPortWrapper wrapper = new MidiPortWrapper(info, mType, i);
            MidiPortWrapper currentWrapper = mCurrentWrapper;
            if (wrapper.equals(currentWrapper)) {
                mCurrentWrapper = null;
                disconnect();
                notifyStatus(null);
            }
            LogTool.i(wrapper + " was removed");
        }
    }

    private void notifyStatus(MidiPortWrapper wrapper) {
        mainHandler.post(() -> {
            if (wrapper == null) {
                onOTGEvent.onOTGDeviceDisconnect();
            } else {
                if (mOutputPort != null) {
                    onOTGEvent.onOTGDeviceConnect(wrapper);
                } else {
                    onOTGEvent.onOTGDeviceDisconnect();
                }
            }
        });
    }

    public MidiPortWrapper getCurrentWrapper() {
        return mCurrentWrapper;
    }

    /**
     * 是否与设备建立连接
     */
    public boolean isConnected() {
        return mCurrentWrapper != null && mOutputPort != null;
    }

    @Override
    public void onDeviceStatusChanged(final MidiDeviceStatus status) {
    }

    @Override
    public void doConnect(MidiPortWrapper wrapper) {
        disconnect();
        final MidiDeviceInfo info = wrapper.getDeviceInfo();
        //新建线程
        HandlerThread ht = new HandlerThread("OpenMidiDevice");
        ht.start();
        if (info != null) {
            mMidiManager.openDevice(info, device -> {
                if (device == null) {
                    LogTool.e("could not open " + info);
                } else {
                    mOpenDevice = device;
                    mOutputPort = device.openOutputPort(wrapper.getPortIndex());
                    if (mOutputPort == null) {
                        LogTool.e("could not open output port for " + info);
                        return;
                    }
                    mOutputPort.connect(mMidiReceiver);
                    notifyStatus(mCurrentWrapper);
                    ht.quitSafely();
                }
            }, new Handler(ht.getLooper()));
            // Don't run the callback on the UI thread because openOutputPort might take a while.
        }
    }

    @Override
    public void disconnect() {
        try {
            if (mOutputPort != null) {
                mOutputPort.disconnect(mMidiReceiver);
            }
            mOutputPort = null;
            if (mOpenDevice != null) {
                mOpenDevice.close();
            }
            mOpenDevice = null;
        } catch (IOException e) {
            LogTool.e("close failed: " + e);
        }
    }
}
