package com.pnlyy.midi_support;

import android.app.PendingIntent;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.hardware.usb.UsbConstants;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbEndpoint;
import android.hardware.usb.UsbInterface;
import android.hardware.usb.UsbManager;
import android.hardware.usb.UsbRequest;
import android.os.Build;
import android.os.Handler;
import android.os.Looper;

import androidx.annotation.NonNull;

import com.pnlyy.midi_support.interfaces.OnMidiData;
import com.pnlyy.midi_support.utils.LogTool;

import java.util.Collection;

/**
 * Android M- midi usb设备支持 by UsbManager
 *
 * @author liben
 * @date 11/30/20
 */
class MidiSupportUnderM implements MidiSupportImpl<UsbDevice> {

    private final Context mContext;

    //已连接的USB音频设备
    private UsbDevice mDevice;

    private UsbEndpoint mUsbEndpointIn = null;
    private UsbEndpoint mUsbEndpointOut = null;
    private UsbInterface mUsbInterface = null;
    private UsbDeviceConnection mUsbDeviceConnection;

    //设备断开事件监听
    private final DetachReceiver detachReceiver = new DetachReceiver();

    @NonNull
    private final OnMidiData onMidiData;

    private final UsbManager usbManager;

    private ReadingThread readingThread;

    @NonNull
    private final MidiDeviceHelper.OnOTGEvent onOTGEvent;

    private static MidiSupportUnderM instance;

    public static void notifyRefresh() {
        if (instance != null) {
            if (!instance.isConnected()) {
                instance.findDeviceAndConnect();
            }
        }
    }

    public MidiSupportUnderM(
            @NonNull Context context,
            @NonNull OnMidiData onMidiData,
            @NonNull MidiDeviceHelper.OnOTGEvent onOTGEvent
    ) {
        instance = this;
        usbManager = (UsbManager) context.getSystemService(Context.USB_SERVICE);
        this.mContext = context.getApplicationContext();

        mContext.registerReceiver(detachReceiver,
                new IntentFilter(UsbManager.ACTION_USB_DEVICE_DETACHED));

        this.onMidiData = onMidiData;
        this.onOTGEvent = onOTGEvent;
    }

    @Override
    public void init() {
        findDeviceAndConnect();
    }

    private void findDeviceAndConnect() {
        try {
            Collection<UsbDevice> devices = usbManager.getDeviceList().values();
            LogTool.i("UsbDevice size: " + devices.size());
            for (UsbDevice device : devices) {
                //过滤非音频设备
                if (!isSupport(device)) {
                    LogTool.i("pass device: " + device);
                    continue;
                }
                doConnect(device);
                break;
            }
        } catch (Throwable e) {
            LogTool.e(e.toString());
            onOTGEvent.onOTGDeviceError(e);
        }
    }

    private boolean isSupport(UsbDevice device) {
        int deviceClass = device.getDeviceClass();
        if ((deviceClass == UsbConstants.USB_CLASS_AUDIO
                && device.getDeviceSubclass() == UsbConstants.USB_CLASS_HID)
                || deviceClass == UsbConstants.USB_CLASS_PER_INTERFACE
                || deviceClass == UsbConstants.USB_CLASS_MISC
                || deviceClass == UsbConstants.USB_CLASS_VENDOR_SPEC) {
            return true;
        }
        int interfaceCount = device.getInterfaceCount();
        for (int i = 0; i < interfaceCount; i++) {
            UsbInterface ui = device.getInterface(i);
            if (ui.getInterfaceClass() == UsbConstants.USB_CLASS_AUDIO) {
                return true;
            }
        }
        return false;
        //|| device.getVendorId() == 1177 || device.getVendorId() == 1410
        //|| device.getVendorId() == 2045 || device.getVendorId() == 1054;
    }

    private void usbRequest(UsbDevice device) {
        LogTool.i("请求连接：" + device);
        String requestAction = getRequestAction();
        //请求结果接收
        mContext.registerReceiver(new RequestReceiver(), new IntentFilter(requestAction));

        //发起请求
        PendingIntent requestIntent = PendingIntent.getBroadcast(
                mContext, 0,
                new Intent(requestAction), 0
        );
        usbManager.requestPermission(device, requestIntent);
    }

    private String getRequestAction() {
        return mContext.getPackageName() + ".USB_PERMISSION";
    }

    private void initEndPoints(UsbDevice device) {
        int interfaceCount = device.getInterfaceCount();
        for (int i = 0; i < interfaceCount; i++) {
            UsbInterface usbInterface = device.getInterface(i);
            if (usbInterface.getInterfaceClass() == UsbConstants.USB_CLASS_AUDIO) {
                mUsbInterface = usbInterface;
                for (int j = 0; j < usbInterface.getEndpointCount(); j++) {
                    UsbEndpoint endpoint = usbInterface.getEndpoint(j);
                    if (endpoint.getType() == UsbConstants.USB_ENDPOINT_XFER_BULK) {
                        if (endpoint.getDirection() == UsbConstants.USB_DIR_IN) {
                            mUsbEndpointIn = endpoint;
                        } else if (endpoint.getDirection() == UsbConstants.USB_DIR_OUT) {
                            mUsbEndpointOut = endpoint;
                        }
                    }
                }
            }
        }
    }

    private boolean check() {
        if (this.mUsbInterface == null) {
            LogTool.i("No MIDI connection!\n");
            return false;
        }
        mUsbDeviceConnection = usbManager.openDevice(mDevice);
        if (mUsbDeviceConnection == null) {
            LogTool.i("Failed to open MIDI device!");
            return false;
        } else if (!mUsbDeviceConnection.claimInterface(mUsbInterface, true)) {
            LogTool.i("Could not claim MIDI interface!\n");
            return false;
        } else if (this.mUsbEndpointIn != null) {
            UsbRequest req = new UsbRequest();
            if (req.initialize(mUsbDeviceConnection, mUsbEndpointIn)) {
                LogTool.i("MIDI initialized!\n");
                return true;
            }
            LogTool.i("Failed to initialize UsbRequest");
            return false;
        } else {
            LogTool.i("mUsbEndpointIn is null");
            return false;
        }
    }

    @Override
    public void doConnect(UsbDevice device) {
        //发送请求
        usbRequest(device);
    }

    @Override
    public void disconnect() {
        if (readingThread != null) {
            readingThread.running = false;
            readingThread = null;
        }
        if (mUsbDeviceConnection != null) {
            try {
                mUsbDeviceConnection.close();
            } catch (Exception e) {
                LogTool.e(e);
            }
            mUsbDeviceConnection = null;
        }
    }

    @Override
    public void onDestroy() {
        disconnect();
        mContext.unregisterReceiver(detachReceiver);
        instance = null;
    }

    @Override
    public boolean isConnected() {
        return mDevice != null;
    }

    @Override
    public void reconnect() {
        disconnect();
        findDeviceAndConnect();
    }

    class RequestReceiver extends BroadcastReceiver {
        @Override
        public void onReceive(Context context, Intent intent) {
            mDevice = intent.getParcelableExtra(UsbManager.EXTRA_DEVICE);
            if (!intent.getBooleanExtra(UsbManager.EXTRA_PERMISSION_GRANTED, false)) {
                LogTool.i("USB permission denied for device " + mDevice);
            } else if (mDevice != null) {
                LogTool.i("USB Permission granted!");
                initEndPoints(mDevice);
                if (check()) {
                    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
                        onOTGEvent.onOTGDeviceConnect(mDevice.getProductName());
                    } else {
                        onOTGEvent.onOTGDeviceConnect(mDevice.getDeviceName());
                    }
                    readingThread = new ReadingThread(onMidiData, mUsbEndpointIn, mUsbDeviceConnection);
                    //Android 10 直接start ANR???
                    new Handler(Looper.getMainLooper()).postDelayed(() ->
                            readingThread.start(), 300);
                    LogTool.i("线程开启");
                } else {
                    LogTool.i("建立连接失败：" + mDevice);
                }
            }
            mContext.unregisterReceiver(this);
        }
    }

    /**
     * 设备断开事件监听
     */
    class DetachReceiver extends BroadcastReceiver {
        @Override
        public void onReceive(Context context, Intent intent) {
            if (intent != null) {
                UsbDevice device = intent.getParcelableExtra(UsbManager.EXTRA_DEVICE);
                if (device != null && mDevice != null
                        && device.getDeviceName().equals(mDevice.getDeviceName())) {
                    LogTool.i("OTG 设备断开连接：" + device);
                    disconnect();
                    mDevice = null;
                    onOTGEvent.onOTGDeviceDisconnect();
                }
            }
        }
    }

    static class ReadingThread extends Thread {
        private final OnMidiData onMidiData;
        private final UsbEndpoint endpoint;
        private final UsbDeviceConnection connection;

        boolean running = false;
        Handler handler = new Handler(Looper.getMainLooper());

        public ReadingThread(OnMidiData onMidiData, UsbEndpoint endpoint, UsbDeviceConnection connection) {
            this.onMidiData = onMidiData;
            this.endpoint = endpoint;
            this.connection = connection;
        }

        @Override
        public void run() {
            running = true;
            //read buffer
            int bufferSize = 64;
            byte[] buffer = new byte[bufferSize];
            byte[] temp = new byte[bufferSize * 2];
            byte[] dataArray = new byte[bufferSize * 2];
            int lastSurplus = 0;
            while (running) {
                try {
                    int readCount = connection.bulkTransfer(endpoint, buffer, bufferSize, 0);
                    if (readCount > 0) {
                        //buffer -> temp
                        System.arraycopy(buffer, 0, temp, lastSurplus, readCount);
                        lastSurplus += readCount;
                        if (lastSurplus >= 4) {
                            //取4倍数
                            int end = lastSurplus / 4 * 4;
                            //temp -> dataArray [0, end]
                            System.arraycopy(temp, 0, dataArray, 0, end);
                            //剩余
                            int s = lastSurplus - end;
                            if (s > 0) {
                                System.arraycopy(temp, end, temp, 0, s);
                                lastSurplus = s;
                            } else {
                                lastSurplus = 0;
                            }
                            int dataCount = readCount / 4;
                            int i = 0;
                            for (int a = 0; a < dataCount; a++) {
                                int type = dataArray[i] & 0xff;
                                if (type >= 8 && type < 14) {
                                    //type
                                    byte t = (byte) (dataArray[i++] << 4 & 0xff);
                                    byte ch = dataArray[i++];
                                    byte key = dataArray[i++];
                                    byte d4 = dataArray[i++];
                                    handler.post(() -> {
                                        onMidiData.onData(new byte[]{ch, t, key, d4},
                                                0, 4, System.currentTimeMillis());
                                    });
                                }
                            }
                        }
                    }
                } catch (Exception e) {
                    LogTool.e(e.toString());
                    break;
                }
            }
            LogTool.i("Reading线程结束");
        }
    }
}
