
#import "MIDIDevice.h"
#import "Obj_Logging.h"

//Singleton Instance
static MIDIDevice *s_MIDIDevice = nil;
static dispatch_queue_t s_queueSendMIDI = nil;

//MIDI Notify Callback
static void MedeliMIDINotifyProc(const MIDINotification *message, void *refCon)
{
    MIDIDevice *thisMIDI = (__bridge MIDIDevice *)refCon;
    [thisMIDI MIDINotify:message];
}

//MIDI Receiving Callback
static void MedeliMIDIReadProc(const MIDIPacketList *pktlist,
                               void *readProcRefCon, void *srcConnRefCon)
{
    MIDIDevice *thisMIDI = (__bridge MIDIDevice *)readProcRefCon;
    if (thisMIDI.delegate) [thisMIDI.delegate MIDIReceived:pktlist];
}

@implementation MIDIDevice
@synthesize delegate;

+ (MIDIDevice *)MIDIDevice {
    @synchronized(self)
    {
        if (s_MIDIDevice == nil) {
            s_MIDIDevice = [[self alloc] init];
        }
    }
    return s_MIDIDevice;
}

+ (MIDIDevice *)startMIDIMonitor:(id<MIDIProcessDelegate>)deg {
    MIDIDevice *dev = [MIDIDevice MIDIDevice];
    [dev setDelegate:deg];
    return dev;
}

- (id)init {
    if ((self = [super init])) {
        sources = [[NSMutableArray alloc] init];
        destinations = [[NSMutableArray alloc] init];

        lastNotificationID = kMIDIMsgIOError;
        restartTimer = nil;

        // Create MIDI Client
        MIDIClientCreate((CFStringRef)@"MIDI Client", MedeliMIDINotifyProc, (__bridge void *)(self), &client);

        // Create MIDI Output Port
        MIDIOutputPortCreate(client, (CFStringRef)@"MIDI Output Port", &outputPort);

        // Create MIDI Input Port
        MIDIInputPortCreate(client, (CFStringRef)@"MIDIMonitor Input Port", MedeliMIDIReadProc, (__bridge void *)self, &inputPort);

        // force restart MIDI device
        OSStatus error = MIDIRestart();
        NSAssert(noErr == error, @"MIDIRestart() fail");

        [self scanExistingDevices];
    }
    s_queueSendMIDI = dispatch_queue_create("logSendMIDI", NULL);
    return self;
}

- (void)scanExistingDevices {
    const ItemCount numberOfDestinations = MIDIGetNumberOfDestinations();
    for (ItemCount index = 0; index < numberOfDestinations; ++index) {
        bool isVirtual = IsNetworkSession(MIDIGetSource(index));
        if (!isVirtual){
            [self connectDestination:MIDIGetDestination(index)];
        }
    }

    const ItemCount numberOfSources = MIDIGetNumberOfSources();
    for (ItemCount index = 0; index < numberOfSources; ++index) {
        
        bool isVirtual = IsNetworkSession(MIDIGetSource(index));
        if (!isVirtual){
            [self connectSource:MIDIGetSource(index)];
        }
        
    }
}

- (BOOL)isConnect {
    
    const ItemCount numberOfDestinations = MIDIGetNumberOfDestinations();
    const ItemCount numberOfSources = MIDIGetNumberOfSources();
    
    for (ItemCount index = 0; index < numberOfDestinations; ++index) {
        bool isVirtual = IsNetworkSession(MIDIGetSource(index));
        if (!isVirtual){
            return YES;
        }
    }
    
    for (ItemCount index = 0; index < numberOfSources; ++index) {
        bool isVirtual = IsNetworkSession(MIDIGetSource(index));
        if (!isVirtual){
            return YES;
        }
    }
    
    return NO;
}

// 添加虚拟MIDI检测方法
static BOOL IsNetworkSession(MIDIEndpointRef ref)
{
    MIDIEntityRef entity = 0;
    MIDIEndpointGetEntity(ref, &entity);

    BOOL hasMidiRtpKey = NO;
    CFPropertyListRef properties = nil;
    OSStatus s = MIDIObjectGetProperties(entity, &properties, true);
    if (!s) {
        NSDictionary *dictionary = (__bridge NSDictionary *)properties;
        hasMidiRtpKey = [dictionary valueForKey:@"apple.midirtp.session"] != nil;
        CFRelease(properties);
    }

    return hasMidiRtpKey;
}



// 连接
- (void)connectDestination:(MIDIEndpointRef)endpoint {
    [delegate MIDI:self destinationAdded:endpoint];
}

// 断开连接
- (void)disconnectDestination:(MIDIEndpointRef)endpoint {
    [delegate MIDI:self destinationRemoved:endpoint];
}

// 连接源
- (void)connectSource:(MIDIEndpointRef)endpoint {
    MIDIPortConnectSource(inputPort, endpoint, (void *)endpoint);
    [delegate MIDI:self sourceAdded:endpoint];
}

// 断开连接源
- (void)disconnectSource:(MIDIEndpointRef)endpoint {
    MIDIPortDisconnectSource(inputPort, endpoint);
    [delegate MIDI:self sourceRemoved:endpoint];
}

- (void)MIDINotifyAdd:(const MIDIObjectAddRemoveNotification *)notification {
    if (notification->childType == kMIDIObjectType_Destination) [self connectDestination:(MIDIEndpointRef)notification->child];
    else if (notification->childType == kMIDIObjectType_Source) [self connectSource:(MIDIEndpointRef)notification->child];
}

- (void)MIDINotifyRemove:(const MIDIObjectAddRemoveNotification *)notification {
    if (notification->childType == kMIDIObjectType_Destination) [self disconnectDestination:(MIDIEndpointRef)notification->child];
    else if (notification->childType == kMIDIObjectType_Source) [self disconnectSource:(MIDIEndpointRef)notification->child];
}

- (void)MIDINotify:(const MIDINotification *)notification {
    switch (notification->messageID) {
        case kMIDIMsgObjectAdded:
            [self MIDINotifyAdd:(const MIDIObjectAddRemoveNotification *)notification];
            break;
        case kMIDIMsgObjectRemoved:
            [self MIDINotifyRemove:(const MIDIObjectAddRemoveNotification *)notification];
            break;
        case kMIDIMsgSetupChanged:
            if (kMIDIMsgPropertyChanged == lastNotificationID) {
                if (restartTimer) {
                    //[restartTimer invalidate];
                    /*启动的timer为非repeat的timer，到时间后会自动invalidate，所以无需手动invalidate，added by sandy*/
                    restartTimer = nil;
                }
                restartTimer = [NSTimer scheduledTimerWithTimeInterval:2.0f target:self selector:@selector(scanExistingDevices) userInfo:nil repeats:NO];
            }
            NSLog(@"kMIDIMsgSetupChanged");
            break;
        case kMIDIMsgPropertyChanged:
            NSLog(@"kMIDIMsgPropertyChanged");
            break;
        case kMIDIMsgThruConnectionsChanged:
            NSLog(@"kMIDIMsgThruConnectionsChanged");
            break;
        case kMIDIMsgSerialPortOwnerChanged:
            NSLog(@"kMIDIMsgSerialPortOwnerChanged");
            break;
        case kMIDIMsgIOError:
            NSLog(@"kMIDIMsgIOError");
            break;
    }
    lastNotificationID = notification->messageID;
}

- (void)sendPacketList:(const MIDIPacketList *)packetList {
    if (0 == MIDIGetNumberOfDestinations()) {
        return;
        NSLog(@"send packetlist fail because number of destinations is zero");
    }
    for (ItemCount index = 0; index < MIDIGetNumberOfDestinations(); ++index) {
        // Get destination endpoint
        MIDIEndpointRef outputEndpoint = MIDIGetDestination(index);

        if (outputEndpoint) {
            OSStatus s = MIDISend(outputPort, outputEndpoint, packetList);
            NSAssert(noErr == s, @"sendPacketList(), MIDISend() fail");
        } else {
            NSLog(@"Can't get destination %ld", index);
        }
    }
}

- (void)sendMIDIMsg:(const UInt8 *)data Length:(UInt32)size {
    assert(size < 65536);
    NSString *sendStr = @"Sending:";
    for (int i = 0; i < size; i++) {
        sendStr = [NSString stringWithFormat:@"%@ %.2X", sendStr, data[i]];
    }
    NSLog(@"%@", sendStr);
//    objc_log_NSString(LOG_DEBUG, sendStr);
    dispatch_sync(s_queueSendMIDI, ^
    {
        Byte packetBuffer[size + 100];

        MIDIPacketList *packetList = (MIDIPacketList *)packetBuffer;
        MIDIPacket *packet = MIDIPacketListInit(packetList);

        packet = MIDIPacketListAdd(packetList, sizeof(packetBuffer), packet, 0, size, data);

        [self sendPacketList:packetList];
    });
}

@end
